@@ -77,18 +77,6 @@ static void condition_wait(Display *dpy, xcondition_t cv)
 	_XGetXCBBuffer(dpy);
 }
 
-static void handle_event(Display *dpy, xcb_generic_event_t *e)
-{
-	if(!e)
-		_XIOError(dpy);
-	dpy->last_request_read = e->full_sequence;
-	if(e->response_type == X_Error)
-		_XError(dpy, (xError *) e);
-	else
-		_XEnq(dpy, (xEvent *) e);
-	free(e);
-}
-
 static void call_handlers(Display *dpy, xcb_generic_reply_t *buf)
 {
 	_XAsyncHandler *async, *next;
@@ -102,68 +90,85 @@ static void call_handlers(Display *dpy, xcb_generic_reply_t *buf)
 	    _XError(dpy, (xError *) buf);
 }
 
+static xcb_generic_event_t * wait_or_poll_for_event(Display *dpy, int wait)
+{
+	xcb_connection_t *c = dpy->xcb->connection;
+	xcb_generic_event_t *event;
+	if(wait)
+	{
+		UnlockDisplay(dpy);
+		event = xcb_wait_for_event(c);
+		LockDisplay(dpy);
+	}
+	else
+		event = xcb_poll_for_event(c);
+	return event;
+}
+
 static void process_responses(Display *dpy, int wait_for_first_event, xcb_generic_error_t **current_error, unsigned int current_request)
 {
 	void *reply;
 	xcb_generic_event_t *event = dpy->xcb->next_event;
 	xcb_generic_error_t *error;
-	PendingRequest *req;
 	xcb_connection_t *c = dpy->xcb->connection;
 	if(!event && dpy->xcb->event_owner == XlibOwnsEventQueue)
-	{
-		if(wait_for_first_event)
-		{
-			UnlockDisplay(dpy);
-			event = xcb_wait_for_event(c);
-			LockDisplay(dpy);
-		}
-		else
-			event = xcb_poll_for_event(c);
-	}
+		event = wait_or_poll_for_event(dpy, wait_for_first_event);
 
 	while(1)
 	{
-		req = dpy->xcb->pending_requests;
+		PendingRequest *req = dpy->xcb->pending_requests;
 		assert(!(req && current_request && !XCB_SEQUENCE_COMPARE(req->sequence, <=, current_request)));
 		if(event && (!req || XCB_SEQUENCE_COMPARE(event->full_sequence, <=, req->sequence)))
 		{
-			if(current_error && event->response_type == 0 && event->full_sequence == current_request)
+			dpy->last_request_read = event->full_sequence;
+			if(event->response_type != X_Error)
 			{
+				_XEnq(dpy, (xEvent *) event);
+				wait_for_first_event = 0;
+			}
+			else if(current_error && event->full_sequence == current_request)
+			{
+				/* This can only occur when called from
+				 * _XReply, which doesn't need a new event. */
 				*current_error = (xcb_generic_error_t *) event;
 				event = 0;
 				break;
 			}
-			handle_event(dpy, event);
-			event = xcb_poll_for_event(c);
+			else
+				_XError(dpy, (xError *) event);
+			free(event);
+			event = wait_or_poll_for_event(dpy, wait_for_first_event);
 		}
 		else if(req && req->waiters != -1)
 		{
 			if(req->sequence == current_request)
 				break;
-			if(!current_request && !(wait_for_first_event && !dpy->head))
+			if(!current_request && !wait_for_first_event)
 				break;
 			dpy->xcb->next_event = event;
 			req->waiters++;
 			assert(req->waiters > 0);
 			condition_wait(dpy, &req->condition);
-			if(--req->waiters == 0)
-				free(req);
+			--req->waiters;
 			event = dpy->xcb->next_event;
 		}
 		else if(req && xcb_poll_for_reply(dpy->xcb->connection, req->sequence, &reply, &error))
 		{
-			dpy->xcb->pending_requests = req->next;
-			if(!dpy->xcb->pending_requests)
-				dpy->xcb->pending_requests_tail = &dpy->xcb->pending_requests;
+			unsigned int sequence = req->sequence;
 			if(!reply)
+			{
+				dpy->xcb->pending_requests = req->next;
+				if(!dpy->xcb->pending_requests)
+					dpy->xcb->pending_requests_tail = &dpy->xcb->pending_requests;
+				free(req);
 				reply = error;
+			}
 			if(reply)
 			{
-				dpy->last_request_read = req->sequence;
+				dpy->last_request_read = sequence;
 				call_handlers(dpy, reply);
+				free(reply);
 			}
-			free(req);
-			free(reply);
 		}
 		else
 			break;
@@ -175,11 +180,13 @@ static void process_responses(Display *dpy, int wait_for_first_event, xcb_generi
 		_XIOError(dpy);
 
 	assert_sequence_less(dpy->last_request_read, dpy->request);
-	assert(!wait_for_first_event || dpy->head);
+	assert(!wait_for_first_event);
 }
 
 int _XEventsQueued(Display *dpy, int mode)
 {
+	if(dpy->flags & XlibDisplayIOError)
+		return 0;
 	if(dpy->xcb->event_owner != XlibOwnsEventQueue)
 		return 0;
 
@@ -196,6 +203,8 @@ int _XEventsQueued(Display *dpy, int mode)
  */
 void _XReadEvents(Display *dpy)
 {
+	if(dpy->flags & XlibDisplayIOError)
+		return;
 	_XSend(dpy, 0, 0);
 	if(dpy->xcb->event_owner != XlibOwnsEventQueue)
 		return;
@@ -213,6 +222,8 @@ void _XReadEvents(Display *dpy)
 void _XSend(Display *dpy, const char *data, long size)
 {
 	xcb_connection_t *c = dpy->xcb->connection;
+	if(dpy->flags & XlibDisplayIOError)
+		return;
 
 	assert(!dpy->xcb->request_extra);
 	dpy->xcb->request_extra = data;
@@ -271,9 +282,11 @@ XID _XAllocID(Display *dpy)
 	XID ret = dpy->xcb->next_xid;
 	dpy->xcb->next_xid = 0;
 
-	assert(!(dpy->flags & XlibDisplayPrivSync));
-	dpy->savedsynchandler = dpy->synchandler;
-	dpy->flags |= XlibDisplayPrivSync;
+	if(!(dpy->flags & XlibDisplayPrivSync))
+	{
+		dpy->savedsynchandler = dpy->synchandler;
+		dpy->flags |= XlibDisplayPrivSync;
+	}
 	dpy->synchandler = _XIDHandler;
 	return ret;
 }
@@ -323,17 +336,6 @@ static PendingRequest * insert_pending_request(Display *dpy)
 	return *cur;
 }
 
-static void remove_pending_request(Display *dpy, PendingRequest *node)
-{
-	PendingRequest **cur = &dpy->xcb->pending_requests;
-	while(*cur && *cur != node)
-		cur = &((*cur)->next);
-	if(*cur == node)
-		*cur = node->next;
-	if(!dpy->xcb->pending_requests)
-		dpy->xcb->pending_requests_tail = &dpy->xcb->pending_requests;
-}
-
 /*
  * _XReply - Wait for a reply packet and copy its contents into the
  * specified rep.
@@ -349,12 +351,16 @@ Status _XReply(Display *dpy, xReply *rep, int extra, Bool discard)
 
 	assert(!dpy->xcb->reply_data);
 
+	if(dpy->flags & XlibDisplayIOError)
+		return 0;
+
 	/* Internals of UnlockDisplay done by hand here, so that we can
 	   insert_pending_request *after* we _XPutXCBBuffer, but before we
 	   unlock the display. */
 	_XPutXCBBuffer(dpy);
 	current = insert_pending_request(dpy);
-	xcb_xlib_unlock(dpy->xcb->connection);
+	if(!dpy->lock || dpy->lock->locking_level == 0)
+		xcb_xlib_unlock(dpy->xcb->connection);
 	if(dpy->xcb->lock_fns.unlock_display)
 		dpy->xcb->lock_fns.unlock_display(dpy);
 	reply = xcb_wait_for_reply(c, current->sequence, &error);
@@ -363,13 +369,11 @@ Status _XReply(Display *dpy, xReply *rep, int extra, Bool discard)
 	check_internal_connections(dpy);
 	process_responses(dpy, 0, &error, current->sequence);
 
-	remove_pending_request(dpy, current);
 	if(current->waiters)
 	{ /* The ConditionBroadcast macro contains an if; braces needed here. */
 		ConditionBroadcast(dpy, &current->condition);
 	}
-	else
-		free(current);
+	--current->waiters;
 
 	if(error)
 	{
